@@ -1,6 +1,6 @@
 /***************************************************************************
     begin       : Wed Jan 20 2010
-    copyright   : (C) 2010 by Martin Preuss
+    copyright   : (C) 2019 by Martin Preuss
     email       : martin@libchipcard.de
 
  ***************************************************************************
@@ -41,6 +41,7 @@
 
 #include <assert.h>
 #include <ctype.h>
+#include <strings.h>
 
 
 GWEN_INHERIT_FUNCTIONS(GWEN_DIALOG)
@@ -50,7 +51,8 @@ GWEN_LIST2_FUNCTIONS(GWEN_DIALOG, GWEN_Dialog)
 
 
 
-GWEN_DIALOG *GWEN_Dialog_new(const char *dialogId) {
+GWEN_DIALOG *GWEN_Dialog_new(const char *dialogId)
+{
   GWEN_DIALOG *dlg;
   int rv;
   GWEN_DB_NODE *db=NULL;
@@ -85,7 +87,55 @@ GWEN_DIALOG *GWEN_Dialog_new(const char *dialogId) {
 
 
 
-void GWEN_Dialog_free(GWEN_DIALOG *dlg) {
+GWEN_DIALOG *GWEN_Dialog_CreateAndLoadWithPath(const char *dialogId,
+                                               const char *pmLibName,
+                                               const char *pmDataDir,
+                                               const char *fileName)
+{
+  GWEN_DIALOG *dlg;
+  GWEN_BUFFER *fbuf;
+  int rv;
+
+  assert(dialogId);
+  assert(pmLibName);
+  assert(pmDataDir);
+  assert(fileName);
+
+  dlg=GWEN_Dialog_new(dialogId);
+  if (dlg==NULL) {
+    DBG_ERROR(GWEN_LOGDOMAIN, "Could not create dialog");
+    return NULL;
+  }
+
+  /* get path of dialog description file */
+  fbuf=GWEN_Buffer_new(0, 256, 0, 1);
+  rv=GWEN_PathManager_FindFile(pmLibName, pmDataDir, fileName, fbuf);
+  if (rv<0) {
+    DBG_ERROR(GWEN_LOGDOMAIN, "Dialog description file not found (%d).", rv);
+    GWEN_Buffer_free(fbuf);
+    GWEN_Dialog_free(dlg);
+    return NULL;
+  }
+
+  /* read dialog from dialog description file */
+  rv=GWEN_Dialog_ReadXmlFile(dlg, GWEN_Buffer_GetStart(fbuf));
+  if (rv<0) {
+    DBG_INFO(GWEN_LOGDOMAIN, "here (%d).", rv);
+    GWEN_Buffer_free(fbuf);
+    GWEN_Dialog_free(dlg);
+    return NULL;
+  }
+  DBG_INFO(GWEN_LOGDOMAIN, "Dialog \"%s\" loaded from \"%s\"", dialogId, GWEN_Buffer_GetStart(fbuf));
+  GWEN_Buffer_free(fbuf);
+
+  return dlg;
+}
+
+
+
+
+void GWEN_Dialog_free(GWEN_DIALOG *dlg)
+{
   if (dlg) {
     assert(dlg->refCount);
 
@@ -98,7 +148,7 @@ void GWEN_Dialog_free(GWEN_DIALOG *dlg) {
       /* try to save preferences */
       rv=GWEN_Gui_WriteDialogPrefs(dlg->dialogId, dlg->dbPreferences);
       if (rv<0) {
-	DBG_WARN(GWEN_LOGDOMAIN, "Could not write dialog preferences (%d)", rv);
+        DBG_WARN(GWEN_LOGDOMAIN, "Could not write dialog preferences (%d)", rv);
       }
 
       GWEN_INHERIT_FINI(GWEN_DIALOG, dlg);
@@ -119,7 +169,8 @@ void GWEN_Dialog_free(GWEN_DIALOG *dlg) {
 
 
 
-GWEN_DIALOG *GWEN_Dialog_GetParentDialog(const GWEN_DIALOG *dlg) {
+GWEN_DIALOG *GWEN_Dialog_GetParentDialog(const GWEN_DIALOG *dlg)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -128,7 +179,8 @@ GWEN_DIALOG *GWEN_Dialog_GetParentDialog(const GWEN_DIALOG *dlg) {
 
 
 
-const char *GWEN_Dialog_GetId(const GWEN_DIALOG *dlg) {
+const char *GWEN_Dialog_GetId(const GWEN_DIALOG *dlg)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -137,7 +189,8 @@ const char *GWEN_Dialog_GetId(const GWEN_DIALOG *dlg) {
 
 
 
-uint32_t GWEN_Dialog_GetGuiId(const GWEN_DIALOG *dlg) {
+uint32_t GWEN_Dialog_GetGuiId(const GWEN_DIALOG *dlg)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -146,7 +199,8 @@ uint32_t GWEN_Dialog_GetGuiId(const GWEN_DIALOG *dlg) {
 
 
 
-void GWEN_Dialog_SetGuiId(GWEN_DIALOG *dlg, uint32_t guiid) {
+void GWEN_Dialog_SetGuiId(GWEN_DIALOG *dlg, uint32_t guiid)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -155,18 +209,22 @@ void GWEN_Dialog_SetGuiId(GWEN_DIALOG *dlg, uint32_t guiid) {
 
 
 
-void GWEN_Dialog_SetI18nDomain(GWEN_DIALOG *dlg, const char *s) {
+void GWEN_Dialog_SetI18nDomain(GWEN_DIALOG *dlg, const char *s)
+{
   assert(dlg);
   assert(dlg->refCount);
 
   free(dlg->i18nDomain);
-  if (s) dlg->i18nDomain=strdup(s);
-  else dlg->i18nDomain=strdup(PACKAGE);
+  if (s)
+    dlg->i18nDomain=strdup(s);
+  else
+    dlg->i18nDomain=strdup(PACKAGE);
 }
 
 
 
-const char *GWEN_Dialog_GetI18nDomain(const GWEN_DIALOG *dlg) {
+const char *GWEN_Dialog_GetI18nDomain(const GWEN_DIALOG *dlg)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -175,7 +233,8 @@ const char *GWEN_Dialog_GetI18nDomain(const GWEN_DIALOG *dlg) {
 
 
 
-const char *GWEN_Dialog_TranslateString(const GWEN_DIALOG *dlg, const char *s) {
+const char *GWEN_Dialog_TranslateString(const GWEN_DIALOG *dlg, const char *s)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -184,7 +243,8 @@ const char *GWEN_Dialog_TranslateString(const GWEN_DIALOG *dlg, const char *s) {
 
 
 
-void GWEN_Dialog_AddMediaPath(GWEN_DIALOG *dlg, const char *s) {
+void GWEN_Dialog_AddMediaPath(GWEN_DIALOG *dlg, const char *s)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -194,9 +254,10 @@ void GWEN_Dialog_AddMediaPath(GWEN_DIALOG *dlg, const char *s) {
 
 
 void GWEN_Dialog_AddMediaPathsFromPathManager(GWEN_DIALOG *dlg,
-					      const char *destlib,
-					      const char *pathName,
-					      const char *relPath) {
+                                              const char *destlib,
+                                              const char *pathName,
+                                              const char *relPath)
+{
   GWEN_STRINGLIST *sl;
 
   sl=GWEN_PathManager_GetPaths(destlib, pathName);
@@ -207,21 +268,21 @@ void GWEN_Dialog_AddMediaPathsFromPathManager(GWEN_DIALOG *dlg,
       GWEN_BUFFER *tbuf;
 
       tbuf=GWEN_Buffer_new(0, 256, 0, 1);
-      while(se) {
-	const char *s;
+      while (se) {
+        const char *s;
 
-	s=GWEN_StringListEntry_Data(se);
-	assert(s);
-	if (relPath) {
-	  GWEN_Buffer_AppendString(tbuf, s);
-	  GWEN_Buffer_AppendString(tbuf, GWEN_DIR_SEPARATOR_S);
-	  GWEN_Buffer_AppendString(tbuf, relPath);
-	  GWEN_StringList_AppendString(dlg->mediaPaths, GWEN_Buffer_GetStart(tbuf), 0, 1);
-	  GWEN_Buffer_Reset(tbuf);
-	}
-	else
-	  GWEN_StringList_AppendString(dlg->mediaPaths, s, 0, 1);
-	se=GWEN_StringListEntry_Next(se);
+        s=GWEN_StringListEntry_Data(se);
+        assert(s);
+        if (relPath) {
+          GWEN_Buffer_AppendString(tbuf, s);
+          GWEN_Buffer_AppendString(tbuf, GWEN_DIR_SEPARATOR_S);
+          GWEN_Buffer_AppendString(tbuf, relPath);
+          GWEN_StringList_AppendString(dlg->mediaPaths, GWEN_Buffer_GetStart(tbuf), 0, 1);
+          GWEN_Buffer_Reset(tbuf);
+        }
+        else
+          GWEN_StringList_AppendString(dlg->mediaPaths, s, 0, 1);
+        se=GWEN_StringListEntry_Next(se);
       }
       GWEN_Buffer_free(tbuf);
     }
@@ -231,7 +292,8 @@ void GWEN_Dialog_AddMediaPathsFromPathManager(GWEN_DIALOG *dlg,
 
 
 
-GWEN_STRINGLIST *GWEN_Dialog_GetMediaPaths(const GWEN_DIALOG *dlg) {
+GWEN_STRINGLIST *GWEN_Dialog_GetMediaPaths(const GWEN_DIALOG *dlg)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -241,7 +303,8 @@ GWEN_STRINGLIST *GWEN_Dialog_GetMediaPaths(const GWEN_DIALOG *dlg) {
 
 
 GWEN_DIALOG_SIGNALHANDLER GWEN_Dialog_SetSignalHandler(GWEN_DIALOG *dlg,
-						       GWEN_DIALOG_SIGNALHANDLER fn) {
+                                                       GWEN_DIALOG_SIGNALHANDLER fn)
+{
   GWEN_DIALOG_SIGNALHANDLER oh;
 
   assert(dlg);
@@ -256,55 +319,114 @@ GWEN_DIALOG_SIGNALHANDLER GWEN_Dialog_SetSignalHandler(GWEN_DIALOG *dlg,
 
 
 int GWEN_Dialog_EmitSignal(GWEN_DIALOG *dlg,
-			   GWEN_DIALOG_EVENTTYPE t,
-			   const char *sender) {
+                           GWEN_DIALOG_EVENTTYPE t,
+                           const char *sender)
+{
+  return GWEN_Dialog_EmitSignal2(dlg, t, sender, -1, NULL);
+}
+
+
+
+int GWEN_Dialog_EmitSignalToAll(GWEN_DIALOG *dlg,
+                                GWEN_DIALOG_EVENTTYPE t,
+                                const char *sender)
+{
+  return GWEN_Dialog_EmitSignalToAll2(dlg, t, sender, -1, NULL);
+}
+
+
+
+GWEN_DIALOG_SIGNALHANDLER2 GWEN_Dialog_SetSignalHandler2(GWEN_DIALOG *dlg,
+							GWEN_DIALOG_SIGNALHANDLER2 fn)
+{
+  GWEN_DIALOG_SIGNALHANDLER2 oh;
+
   assert(dlg);
   assert(dlg->refCount);
 
-  if (dlg->signalHandler)
+  oh=dlg->signalHandler2;
+  dlg->signalHandler2=fn;
+
+  return oh;
+}
+
+
+
+int GWEN_Dialog_EmitSignal2(GWEN_DIALOG *dlg,
+                            GWEN_DIALOG_EVENTTYPE t,
+			    const char *sender,
+			    int intArg,
+			    const char *stringArg)
+{
+  assert(dlg);
+  assert(dlg->refCount);
+
+  if (dlg->signalHandler2)
+    return (dlg->signalHandler2)(dlg, t, sender, intArg, stringArg);
+  else if (dlg->signalHandler)
     return (dlg->signalHandler)(dlg, t, sender);
   else {
     DBG_WARN(GWEN_LOGDOMAIN, "No signal handler in dialog [%s]",
-	     (dlg->dialogId)?(dlg->dialogId):"-unnamed-");
+             (dlg->dialogId)?(dlg->dialogId):"-unnamed-");
     return GWEN_DialogEvent_ResultNotHandled;
   }
 }
 
 
 
-int GWEN_Dialog_EmitSignalToAll(GWEN_DIALOG *dlg,
-				GWEN_DIALOG_EVENTTYPE t,
-				const char *sender) {
+int GWEN_Dialog_EmitSignalToAll2(GWEN_DIALOG *dlg,
+                                 GWEN_DIALOG_EVENTTYPE t,
+				 const char *sender,
+				 int intArg,
+				 const char *stringArg)
+{
   int rv;
   GWEN_DIALOG *subdlg;
+  int anyDialogHasHandledThisSignal=0;
 
   assert(dlg);
   assert(dlg->refCount);
 
   subdlg=GWEN_Dialog_List_First(dlg->subDialogs);
-  while(subdlg) {
-    rv=GWEN_Dialog_EmitSignalToAll(subdlg, t, sender);
+  while (subdlg) {
+    rv=GWEN_Dialog_EmitSignalToAll2(subdlg, t, sender, intArg, stringArg);
+    if (rv==GWEN_DialogEvent_ResultHandled)
+      anyDialogHasHandledThisSignal=1;
     if (rv!=GWEN_DialogEvent_ResultHandled &&
-	rv!=GWEN_DialogEvent_ResultNotHandled)
+        rv!=GWEN_DialogEvent_ResultNotHandled)
       return rv;
     subdlg=GWEN_Dialog_List_Next(subdlg);
   }
 
-  if (dlg->signalHandler) {
+  if (dlg->signalHandler2) {
+    rv=(dlg->signalHandler2)(dlg, t, sender, intArg, stringArg);
+    if (rv==GWEN_DialogEvent_ResultHandled)
+      anyDialogHasHandledThisSignal=1;
+    if (rv!=GWEN_DialogEvent_ResultHandled &&
+        rv!=GWEN_DialogEvent_ResultNotHandled)
+      return rv;
+  }
+  else if (dlg->signalHandler) {
     rv=(dlg->signalHandler)(dlg, t, sender);
+    if (rv==GWEN_DialogEvent_ResultHandled)
+      anyDialogHasHandledThisSignal=1;
     if (rv!=GWEN_DialogEvent_ResultHandled &&
 	rv!=GWEN_DialogEvent_ResultNotHandled)
       return rv;
   }
 
-  return GWEN_DialogEvent_ResultHandled;
+  if (anyDialogHasHandledThisSignal)
+    return GWEN_DialogEvent_ResultHandled;
+  else
+    return GWEN_DialogEvent_ResultNotHandled;
 }
 
 
 
 int GWEN_Dialog_AddSubDialog(GWEN_DIALOG *dlg,
                              const char *parentName,
-			     GWEN_DIALOG *subdlg) {
+                             GWEN_DIALOG *subdlg)
+{
   GWEN_WIDGET *wparent;
 
   wparent=GWEN_Dialog_FindWidgetByName(dlg, parentName);
@@ -315,11 +437,11 @@ int GWEN_Dialog_AddSubDialog(GWEN_DIALOG *dlg,
     /* move all widgets from the below sub dialog's root widget to the parent dialog */
     subRoot=GWEN_Widget_Tree_GetFirst(subdlg->widgets);
     if (subRoot) {
-      while( (cw=GWEN_Widget_Tree_GetFirstChild(subRoot)) ) {
-	if (subdlg->firstSubWidget==NULL)
-	  subdlg->firstSubWidget=cw;
-	GWEN_Widget_Tree_Del(cw);
-	GWEN_Widget_Tree_AddChild(wparent, cw);
+      while ((cw=GWEN_Widget_Tree_GetFirstChild(subRoot))) {
+        if (subdlg->firstSubWidget==NULL)
+          subdlg->firstSubWidget=cw;
+        GWEN_Widget_Tree_Del(cw);
+        GWEN_Widget_Tree_AddChild(wparent, cw);
       }
     }
 
@@ -342,8 +464,9 @@ int GWEN_Dialog_AddSubDialog(GWEN_DIALOG *dlg,
 
 
 int GWEN_Dialog__ReadXmlWidget(GWEN_DIALOG *dlg,
-			       GWEN_WIDGET *wparent,
-			       GWEN_XMLNODE *node) {
+                               GWEN_WIDGET *wparent,
+                               GWEN_XMLNODE *node)
+{
   GWEN_XMLNODE *n;
   GWEN_WIDGET *w;
   int rv;
@@ -363,7 +486,7 @@ int GWEN_Dialog__ReadXmlWidget(GWEN_DIALOG *dlg,
 
 
   n=GWEN_XMLNode_FindFirstTag(node, "widget", NULL, NULL);
-  while(n) {
+  while (n) {
     int rv;
 
     rv=GWEN_Dialog__ReadXmlWidget(dlg, w, n);
@@ -377,7 +500,8 @@ int GWEN_Dialog__ReadXmlWidget(GWEN_DIALOG *dlg,
 
 
 
-int GWEN_Dialog_ReadXml(GWEN_DIALOG *dlg, GWEN_XMLNODE *node) {
+int GWEN_Dialog_ReadXml(GWEN_DIALOG *dlg, GWEN_XMLNODE *node)
+{
   int rv;
   const char *s;
 
@@ -404,15 +528,16 @@ int GWEN_Dialog_ReadXml(GWEN_DIALOG *dlg, GWEN_XMLNODE *node) {
 
 
 
-int GWEN_Dialog_ReadXmlFile(GWEN_DIALOG *dlg, const char *fname) {
+int GWEN_Dialog_ReadXmlFile(GWEN_DIALOG *dlg, const char *fname)
+{
   GWEN_XMLNODE *n;
   GWEN_XMLNODE *nDialog;
   int rv;
 
   n=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "root");
   rv=GWEN_XML_ReadFile(n, fname,
-		       GWEN_XML_FLAGS_DEFAULT |
-		       GWEN_XML_FLAGS_HANDLE_HEADERS);
+                       GWEN_XML_FLAGS_DEFAULT |
+                       GWEN_XML_FLAGS_HANDLE_HEADERS);
   if (rv<0) {
     DBG_ERROR(GWEN_LOGDOMAIN, "here (%d)", rv);
     GWEN_XMLNode_free(n);
@@ -439,7 +564,8 @@ int GWEN_Dialog_ReadXmlFile(GWEN_DIALOG *dlg, const char *fname) {
 
 
 
-GWEN_WIDGET *GWEN_Dialog_FindWidgetByName(const GWEN_DIALOG *dlg, const char *name) {
+GWEN_WIDGET *GWEN_Dialog_FindWidgetByName(const GWEN_DIALOG *dlg, const char *name)
+{
   GWEN_WIDGET *w;
 
   assert(dlg);
@@ -462,7 +588,7 @@ GWEN_WIDGET *GWEN_Dialog_FindWidgetByName(const GWEN_DIALOG *dlg, const char *na
   }
 #endif
 
-  while(w) {
+  while (w) {
     const char *s;
 
     s=GWEN_Widget_GetName(w);
@@ -476,7 +602,8 @@ GWEN_WIDGET *GWEN_Dialog_FindWidgetByName(const GWEN_DIALOG *dlg, const char *na
 
 
 
-GWEN_WIDGET *GWEN_Dialog_FindWidgetByImplData(const GWEN_DIALOG *dlg, int index, const void *ptr) {
+GWEN_WIDGET *GWEN_Dialog_FindWidgetByImplData(const GWEN_DIALOG *dlg, int index, const void *ptr)
+{
   GWEN_WIDGET *w;
 
   assert(dlg);
@@ -488,7 +615,7 @@ GWEN_WIDGET *GWEN_Dialog_FindWidgetByImplData(const GWEN_DIALOG *dlg, int index,
   else
     w=GWEN_Widget_Tree_GetFirst(dlg->widgets);
 
-  while(w) {
+  while (w) {
     if (ptr==GWEN_Widget_GetImplData(w, index))
       break;
     w=GWEN_Widget_Tree_GetBelow(w);
@@ -499,7 +626,8 @@ GWEN_WIDGET *GWEN_Dialog_FindWidgetByImplData(const GWEN_DIALOG *dlg, int index,
 
 
 
-GWEN_WIDGET_TREE *GWEN_Dialog_GetWidgets(const GWEN_DIALOG *dlg) {
+GWEN_WIDGET_TREE *GWEN_Dialog_GetWidgets(const GWEN_DIALOG *dlg)
+{
   assert(dlg);
   assert(dlg->refCount);
   assert(dlg->widgets);
@@ -510,7 +638,8 @@ GWEN_WIDGET_TREE *GWEN_Dialog_GetWidgets(const GWEN_DIALOG *dlg) {
 
 
 GWEN_DIALOG_SETINTPROPERTY_FN GWEN_Dialog_SetSetIntPropertyFn(GWEN_DIALOG *dlg,
-							      GWEN_DIALOG_SETINTPROPERTY_FN fn) {
+                                                              GWEN_DIALOG_SETINTPROPERTY_FN fn)
+{
   GWEN_DIALOG_SETINTPROPERTY_FN oh;
 
   assert(dlg);
@@ -524,7 +653,8 @@ GWEN_DIALOG_SETINTPROPERTY_FN GWEN_Dialog_SetSetIntPropertyFn(GWEN_DIALOG *dlg,
 
 
 GWEN_DIALOG_GETINTPROPERTY_FN GWEN_Dialog_SetGetIntPropertyFn(GWEN_DIALOG *dlg,
-							      GWEN_DIALOG_GETINTPROPERTY_FN fn) {
+                                                              GWEN_DIALOG_GETINTPROPERTY_FN fn)
+{
   GWEN_DIALOG_GETINTPROPERTY_FN oh;
 
   assert(dlg);
@@ -538,7 +668,8 @@ GWEN_DIALOG_GETINTPROPERTY_FN GWEN_Dialog_SetGetIntPropertyFn(GWEN_DIALOG *dlg,
 
 
 GWEN_DIALOG_SETCHARPROPERTY_FN GWEN_Dialog_SetSetCharPropertyFn(GWEN_DIALOG *dlg,
-								GWEN_DIALOG_SETCHARPROPERTY_FN fn) {
+                                                                GWEN_DIALOG_SETCHARPROPERTY_FN fn)
+{
   GWEN_DIALOG_SETCHARPROPERTY_FN oh;
 
   assert(dlg);
@@ -552,7 +683,8 @@ GWEN_DIALOG_SETCHARPROPERTY_FN GWEN_Dialog_SetSetCharPropertyFn(GWEN_DIALOG *dlg
 
 
 GWEN_DIALOG_GETCHARPROPERTY_FN GWEN_Dialog_SetGetCharPropertyFn(GWEN_DIALOG *dlg,
-								GWEN_DIALOG_GETCHARPROPERTY_FN fn) {
+                                                                GWEN_DIALOG_GETCHARPROPERTY_FN fn)
+{
   GWEN_DIALOG_GETCHARPROPERTY_FN oh;
 
   assert(dlg);
@@ -569,11 +701,12 @@ GWEN_DIALOG_GETCHARPROPERTY_FN GWEN_Dialog_SetGetCharPropertyFn(GWEN_DIALOG *dlg
 
 
 int GWEN_Dialog_SetIntProperty(GWEN_DIALOG *dlg,
-			       const char *name,
-			       GWEN_DIALOG_PROPERTY prop,
-			       int index,
-			       int value,
-			       int doSignal) {
+                               const char *name,
+                               GWEN_DIALOG_PROPERTY prop,
+                               int index,
+                               int value,
+                               int doSignal)
+{
   GWEN_WIDGET *w;
 
   assert(dlg);
@@ -598,10 +731,11 @@ int GWEN_Dialog_SetIntProperty(GWEN_DIALOG *dlg,
 
 
 int GWEN_Dialog_GetIntProperty(GWEN_DIALOG *dlg,
-			       const char *name,
-			       GWEN_DIALOG_PROPERTY prop,
-			       int index,
-			       int defaultProperty) {
+                               const char *name,
+                               GWEN_DIALOG_PROPERTY prop,
+                               int index,
+                               int defaultProperty)
+{
   GWEN_WIDGET *w;
 
   assert(dlg);
@@ -626,11 +760,12 @@ int GWEN_Dialog_GetIntProperty(GWEN_DIALOG *dlg,
 
 
 int GWEN_Dialog_SetCharProperty(GWEN_DIALOG *dlg,
-				const char *name,
-				GWEN_DIALOG_PROPERTY prop,
-				int index,
-				const char *value,
-				int doSignal) {
+                                const char *name,
+                                GWEN_DIALOG_PROPERTY prop,
+                                int index,
+                                const char *value,
+                                int doSignal)
+{
   GWEN_WIDGET *w;
 
   assert(dlg);
@@ -655,10 +790,11 @@ int GWEN_Dialog_SetCharProperty(GWEN_DIALOG *dlg,
 
 
 const char *GWEN_Dialog_GetCharProperty(GWEN_DIALOG *dlg,
-					const char *name,
-					GWEN_DIALOG_PROPERTY prop,
-					int index,
-					const char *defaultProperty) {
+                                        const char *name,
+                                        GWEN_DIALOG_PROPERTY prop,
+                                        int index,
+                                        const char *defaultProperty)
+{
   GWEN_WIDGET *w;
 
   assert(dlg);
@@ -682,7 +818,8 @@ const char *GWEN_Dialog_GetCharProperty(GWEN_DIALOG *dlg,
 
 
 
-int GWEN_Dialog_RemoveWidget(GWEN_DIALOG *dlg, const char *name) {
+int GWEN_Dialog_RemoveWidget(GWEN_DIALOG *dlg, const char *name)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -696,7 +833,8 @@ int GWEN_Dialog_RemoveWidget(GWEN_DIALOG *dlg, const char *name) {
 }
 
 
-uint32_t GWEN_Dialog_GetWidgetFlags(const GWEN_DIALOG *dlg, const char *name) {
+uint32_t GWEN_Dialog_GetWidgetFlags(const GWEN_DIALOG *dlg, const char *name)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -709,7 +847,8 @@ uint32_t GWEN_Dialog_GetWidgetFlags(const GWEN_DIALOG *dlg, const char *name) {
 
 
 
-void GWEN_Dialog_SetWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl) {
+void GWEN_Dialog_SetWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -720,7 +859,8 @@ void GWEN_Dialog_SetWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl)
 
 
 
-void GWEN_Dialog_AddWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl) {
+void GWEN_Dialog_AddWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -731,7 +871,8 @@ void GWEN_Dialog_AddWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl)
 
 
 
-void GWEN_Dialog_SubWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl) {
+void GWEN_Dialog_SubWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -742,7 +883,8 @@ void GWEN_Dialog_SubWidgetFlags(GWEN_DIALOG *dlg, const char *name, uint32_t fl)
 
 
 
-int GWEN_Dialog_GetWidgetColumns(const GWEN_DIALOG *dlg, const char *name) {
+int GWEN_Dialog_GetWidgetColumns(const GWEN_DIALOG *dlg, const char *name)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -755,7 +897,8 @@ int GWEN_Dialog_GetWidgetColumns(const GWEN_DIALOG *dlg, const char *name) {
 
 
 
-void GWEN_Dialog_SetWidgetColumns(GWEN_DIALOG *dlg, const char *name, int i) {
+void GWEN_Dialog_SetWidgetColumns(GWEN_DIALOG *dlg, const char *name, int i)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -766,7 +909,8 @@ void GWEN_Dialog_SetWidgetColumns(GWEN_DIALOG *dlg, const char *name, int i) {
 
 
 
-int GWEN_Dialog_GetWidgetRows(const GWEN_DIALOG *dlg, const char *name) {
+int GWEN_Dialog_GetWidgetRows(const GWEN_DIALOG *dlg, const char *name)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -779,7 +923,8 @@ int GWEN_Dialog_GetWidgetRows(const GWEN_DIALOG *dlg, const char *name) {
 
 
 
-void GWEN_Dialog_SetWidgetRows(GWEN_DIALOG *dlg, const char *name, int i) {
+void GWEN_Dialog_SetWidgetRows(GWEN_DIALOG *dlg, const char *name, int i)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -790,7 +935,8 @@ void GWEN_Dialog_SetWidgetRows(GWEN_DIALOG *dlg, const char *name, int i) {
 
 
 
-void GWEN_Dialog_SetWidgetText(GWEN_DIALOG *dlg, const char *name, const char *t) {
+void GWEN_Dialog_SetWidgetText(GWEN_DIALOG *dlg, const char *name, const char *t)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -801,7 +947,8 @@ void GWEN_Dialog_SetWidgetText(GWEN_DIALOG *dlg, const char *name, const char *t
 
 
 
-const char *GWEN_Dialog_GetWidgetText(const GWEN_DIALOG *dlg, const char *name) {
+const char *GWEN_Dialog_GetWidgetText(const GWEN_DIALOG *dlg, const char *name)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Dialog_FindWidgetByName(dlg, name);
@@ -814,7 +961,8 @@ const char *GWEN_Dialog_GetWidgetText(const GWEN_DIALOG *dlg, const char *name)
 
 
 
-static void GWEN_Dialog_DumpWidget(const GWEN_WIDGET *w, FILE *f, unsigned int indent) {
+static void GWEN_Dialog_DumpWidget(const GWEN_WIDGET *w, FILE *f, unsigned int indent)
+{
   int i;
   const char *s;
   const char *dname=NULL;
@@ -830,7 +978,7 @@ static void GWEN_Dialog_DumpWidget(const GWEN_WIDGET *w, FILE *f, unsigned int i
 
   fprintf(f, "Widget %s: %d [%s]\n", s?s:"unnamed", GWEN_Widget_GetType(w), dname?dname:"no dialog");
   c=GWEN_Widget_Tree_GetFirstChild(w);
-  while(c) {
+  while (c) {
     GWEN_Dialog_DumpWidget(c, f, indent+2);
     c=GWEN_Widget_Tree_GetNext(c);
   }
@@ -838,7 +986,8 @@ static void GWEN_Dialog_DumpWidget(const GWEN_WIDGET *w, FILE *f, unsigned int i
 
 
 
-void GWEN_Dialog_Dump(const GWEN_DIALOG *dlg, FILE *f, unsigned int indent) {
+void GWEN_Dialog_Dump(const GWEN_DIALOG *dlg, FILE *f, unsigned int indent)
+{
   GWEN_WIDGET *w;
 
   w=GWEN_Widget_Tree_GetFirst(dlg->widgets);
@@ -850,7 +999,8 @@ void GWEN_Dialog_Dump(const GWEN_DIALOG *dlg, FILE *f, unsigned int indent) {
 
 
 
-GWEN_DB_NODE *GWEN_Dialog_GetPreferences(const GWEN_DIALOG *dlg) {
+GWEN_DB_NODE *GWEN_Dialog_GetPreferences(const GWEN_DIALOG *dlg)
+{
   assert(dlg);
   assert(dlg->refCount);
 
@@ -860,3 +1010,166 @@ GWEN_DB_NODE *GWEN_Dialog_GetPreferences(const GWEN_DIALOG *dlg) {
 
 
 
+
+void GWEN_Dialog_ListWriteColumnSettings(GWEN_DIALOG *dlg,
+					 const char *widgetName,
+					 const char *variablePrefix,
+					 int maxColumns,
+					 int minColumnSize,
+					 GWEN_DB_NODE *dbPrefs)
+{
+  GWEN_BUFFER *nameBuf;
+  uint32_t posInBuffer;
+  int i;
+
+  nameBuf=GWEN_Buffer_new(0, 256, 0, 1);
+  GWEN_Buffer_AppendString(nameBuf, variablePrefix);
+  posInBuffer=GWEN_Buffer_GetPos(nameBuf);
+
+  /* write column sizes */
+  GWEN_Buffer_AppendString(nameBuf, "columns");
+  GWEN_DB_DeleteVar(dbPrefs, GWEN_Buffer_GetStart(nameBuf));
+  for (i=0; i<maxColumns; i++) {
+    int j;
+
+    j=GWEN_Dialog_GetIntProperty(dlg, widgetName, GWEN_DialogProperty_ColumnWidth, i, -1);
+    if (j<minColumnSize)
+      j=minColumnSize;
+    GWEN_DB_SetIntValue(dbPrefs, GWEN_DB_FLAGS_DEFAULT, GWEN_Buffer_GetStart(nameBuf), j);
+  }
+
+  /* determine column to sort by */
+  GWEN_Buffer_Crop(nameBuf, 0, posInBuffer);
+  GWEN_Buffer_AppendString(nameBuf, "sortbycolumn");
+  GWEN_DB_SetIntValue(dbPrefs, GWEN_DB_FLAGS_OVERWRITE_VARS, GWEN_Buffer_GetStart(nameBuf), -1);
+  for (i=0; i<maxColumns; i++) {
+    int j;
+
+    j=GWEN_Dialog_GetIntProperty(dlg, widgetName, GWEN_DialogProperty_SortDirection, i, GWEN_DialogSortDirection_None);
+    if (j!=GWEN_DialogSortDirection_None) {
+
+      GWEN_DB_SetIntValue(dbPrefs, GWEN_DB_FLAGS_OVERWRITE_VARS, GWEN_Buffer_GetStart(nameBuf), i);
+
+      GWEN_Buffer_Crop(nameBuf, 0, posInBuffer);
+      GWEN_Buffer_AppendString(nameBuf, "sortdir");
+      GWEN_DB_SetIntValue(dbPrefs,
+			  GWEN_DB_FLAGS_OVERWRITE_VARS,
+			  GWEN_Buffer_GetStart(nameBuf),
+			  (j==GWEN_DialogSortDirection_Up)?1:0);
+      break;
+    }
+  }
+  GWEN_Buffer_free(nameBuf);
+}
+
+
+
+void GWEN_Dialog_ListReadColumnSettings(GWEN_DIALOG *dlg,
+					const char *widgetName,
+					const char *variablePrefix,
+					int maxColumns,
+					int minColumnSize,
+					GWEN_DB_NODE *dbPrefs)
+{
+  GWEN_BUFFER *nameBuf;
+  uint32_t posInBuffer;
+  int i;
+  int j;
+
+  nameBuf=GWEN_Buffer_new(0, 256, 0, 1);
+  GWEN_Buffer_AppendString(nameBuf, variablePrefix);
+  posInBuffer=GWEN_Buffer_GetPos(nameBuf);
+
+  /* read column sizes */
+  GWEN_Buffer_AppendString(nameBuf, "columns");
+  for (i=0; i<maxColumns; i++) {
+    j=GWEN_DB_GetIntValue(dbPrefs, GWEN_Buffer_GetStart(nameBuf), i, -1);
+    if (j<minColumnSize)
+      j=minColumnSize;
+    GWEN_Dialog_SetIntProperty(dlg, widgetName, GWEN_DialogProperty_ColumnWidth, i, j, 0);
+  }
+
+  /* determine column to sort by */
+  GWEN_Buffer_Crop(nameBuf, 0, posInBuffer);
+  GWEN_Buffer_AppendString(nameBuf, "sortbycolumn");
+  i=GWEN_DB_GetIntValue(dbPrefs, GWEN_Buffer_GetStart(nameBuf), 0, -1);
+  GWEN_Buffer_Crop(nameBuf, 0, posInBuffer);
+  GWEN_Buffer_AppendString(nameBuf, "sortdir");
+  j=GWEN_DB_GetIntValue(dbPrefs, GWEN_Buffer_GetStart(nameBuf), 0, -1);
+  if (i>=0 && j>=0)
+    GWEN_Dialog_SetIntProperty(dlg, widgetName, GWEN_DialogProperty_SortDirection, i, j, 0);
+  GWEN_Buffer_free(nameBuf);
+}
+
+
+
+char *GWEN_Dialog_ListGetFirstColumnData(GWEN_DIALOG *dlg, const char *widgetName, int row)
+{
+  const char *s;
+
+  s=GWEN_Dialog_GetCharProperty(dlg, widgetName, GWEN_DialogProperty_Value, row, NULL);
+  if (s && *s) {
+    const char *p;
+
+    p=strchr(s, '\t');
+    if (p) {
+      int len;
+
+      len=(int)(p-s);
+      if (len) {
+	char *res;
+
+	res=(char *) malloc(len+1);
+	assert(res);
+	memmove(res, s, len);
+	res[len]=0;
+	return res;
+      }
+    }
+    else
+      /* no tab, use the whole line */
+      return strdup(s);
+  }
+
+  return NULL;
+}
+
+
+
+int GWEN_Dialog_ListGetItemMatchingFirstColumn(GWEN_DIALOG *dlg, const char *widgetName, const char *dataToMatch)
+{
+  if (dataToMatch) {
+    int idx;
+    int lengthOfDataToMatch;
+
+    lengthOfDataToMatch=strlen(dataToMatch);
+    for (idx=0; ; idx++) {
+      const char *s;
+
+      s=GWEN_Dialog_GetCharProperty(dlg, widgetName, GWEN_DialogProperty_Value, idx, NULL);
+      if (s && *s) {
+	const char *p;
+
+	p=strchr(s, '\t');
+	if (p) {
+	  int len;
+
+	  len=(int)(p-s);
+	  if (len && len==lengthOfDataToMatch && strncasecmp(s, dataToMatch, len)==0)
+	    return idx;
+	}
+	else if (strcasecmp(s, dataToMatch)==0)
+	  return idx;
+      }
+      else
+	break;
+    }
+  }
+
+  return -1;
+}
+
+
+
+
+
